//	Vamos Automotive Simulator
//  Copyright (C) 2001--2003 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <vamos/track/Road_Segment.h>
#include <vamos/media/Texture_Image.h>

#include <cmath>
#include <cassert>
#include <iostream>

using namespace Vamos_Geometry;
using namespace Vamos_Track;

//=============================================================================

Kerb::Kerb (const std::vector <Two_Vector>& profile,
			double start, 
			double start_transition_length, 
			double start_transition_width,
			double end, 
			double end_transition_length, 
			double end_transition_width) :
  m_points (profile),
  m_profile (profile),
  m_start (start),
  m_start_transition_length (start_transition_length),
  m_start_transition_width (start_transition_width),
  m_end (end),
  m_end_transition_length (end_transition_length),
  m_end_transition_width (end_transition_width),
  m_full_length (m_end < 0.0)
{
}

void 
Kerb::length (double length)
{
  if (m_full_length)
	{
	  m_end = length;
	}
}

double 
Kerb::width () const
{
  return (m_points.size () > 0) ? (m_points.end () - 1)->x : 0.0;
}

bool 
Kerb::on_kerb (double dist) const
{
  //!!TODO handle skew
  return bool ((dist >= m_start) 
			   && ((dist < m_end) || (m_end < 0.0)));
}

double 
Kerb::elevation (double along, double from_inside)
{
  //!!TODO handle skew
  if (!on_kerb (along) || (from_inside < 0.0) || (from_inside > width ()))
	{
	  return 0.0;
	}
  return m_profile.interpolate (from_inside);
}

double 
Kerb::angle (double along, double from_inside)
{
  if (!on_kerb (along) || (from_inside < 0.0) || (from_inside > width ()))
	{
	  return 0.0;
	}
  Two_Vector norm = m_profile.normal (from_inside);
  return std::atan2 (norm.y, norm.x) - pi / 2.0;
}

const Two_Vector& 
Kerb::point (size_t substrip)
{
  assert (substrip < m_points.size ());
  return m_points [substrip];
}

//=============================================================================
Road_Segment::Road_Segment (double length,
                            double radius,
                            double left_width,
                            double right_width,
                            double left_road_width,
                            double right_road_width)
  : m_length (length),
    m_radius (radius),
    m_start_skew (0.0),
    m_end_skew (0.0),
    mp_left_width (new Linear_Interpolator),
    mp_right_width (new Linear_Interpolator),
    mp_left_road_width (new Linear_Interpolator),
    mp_right_road_width (new Linear_Interpolator),
    mp_left_kerb (0),
    mp_right_kerb (0),
    m_left_wall_height (0.0),
    m_right_wall_height (0.0),
    mp_elevation_curve(0),
    m_start_bank (0.0),
    m_end_bank (0.0),
    m_bank_pivot (0.0),
    mp_bank_curve (new Spline),
    m_start_angle (0.0),
    m_last_segment (false)
{
  mp_left_road_width->load (Two_Vector (0.0, left_road_width));
  mp_right_road_width->load (Two_Vector (0.0, right_road_width));
  mp_left_width->load (Two_Vector (0.0, left_width));
  mp_right_width->load (Two_Vector (0.0, right_width));
}

Road_Segment::~Road_Segment ()
{
  delete mp_bank_curve;
  delete mp_right_road_width;
  delete mp_left_road_width;
  delete mp_right_width;
  delete mp_left_width;

  delete mp_left_kerb;
  delete mp_right_kerb;
}

void
Road_Segment::set_skew (double skew)
{
  m_start_skew = skew;
  m_end_skew = -skew;
}

void
Road_Segment::set_width_point (double distance,
                               double width,
                               Side side)
{
  if (side == LEFT)
    {
      mp_left_width->load (Two_Vector (distance, width));
    }
  else
    {
      mp_right_width->load (Two_Vector (distance, width));
    }
}

void
Road_Segment::set_road_width_point (double distance,
                                    double width,
                                    Side side)
{
  if (side == LEFT)
    {
      mp_left_road_width->load (Two_Vector (distance, width));
    }
  else
    {
      mp_right_road_width->load (Two_Vector (distance, width));
    }
}

void
Road_Segment::set_kerb (Kerb* kerb, Side side)
{
  if (side == LEFT)
    {
      delete mp_left_kerb;
      mp_left_kerb = kerb;
    }
  else
    {
      delete mp_right_kerb;
      mp_right_kerb = kerb;
    }
}

void
Road_Segment::set_wall_height (double left_height, double right_height)
{
  m_left_wall_height = left_height;
  m_right_wall_height = right_height;
}

void
Road_Segment::set_elevation_point (double distance, double elevation)
{
  m_elevation_points.push_back (Two_Vector (distance, elevation));
}

void
Road_Segment::set_bank_point (double distance, double bank)
{
  m_bank_points.push_back (Two_Vector (distance, bank));
}

void 
Road_Segment::build_elevation (Spline* elevation,
							   double start_distance)
{
  mp_elevation_curve = elevation;
  for (std::vector <Two_Vector>::iterator it = m_elevation_points.begin ();
	   it != m_elevation_points.end ();
	   it++)
	{
	  mp_elevation_curve->load (*it + Two_Vector (start_distance, 0.0));
	}

  if (m_last_segment)
	{
	  // Remove any elevation points near or greater than m_length.
	  mp_elevation_curve->remove_greater (start_distance + m_length - 10.0);
	  // Force the elevation curve to return to zero.
	  mp_elevation_curve->load (Two_Vector (start_distance + m_length, 0.0)); 
	}
}

double 
Road_Segment::elevation (double along, double from_center) const
{
  assert (mp_elevation_curve != 0);
  double elev = mp_elevation_curve->interpolate (along + m_start_distance)
	+ bank_height (along, from_center);

  // Add the kerb's height.
  double diff = from_center - left_road_width (along);
  if (mp_left_kerb)
    elev += mp_left_kerb->elevation (along, diff);
  diff = -from_center - right_road_width (along);
  if (mp_right_kerb)
    elev += mp_right_kerb->elevation (along, diff);

  return elev;
}

double 
Road_Segment::angle (double along) const
{
  return m_start_angle + arc () * along / m_length;
}

Three_Vector 
Road_Segment::barrier_normal (double along, double from_center) const
{
  double this_angle = angle (along);
  if (from_center > 0.0)
	{
	  return Three_Vector (sin (this_angle), -cos (this_angle), 0.0);
	}
  return Three_Vector (-sin (this_angle), cos (this_angle), 0.0);
}

double 
Road_Segment::bank (double along) const
{
  return deg_to_rad (mp_bank_curve->interpolate (along));
}

double 
Road_Segment::bank_height (double along, double from_center) const
{
  return (m_bank_pivot - from_center) * sin (bank (along));
}


Three_Vector 
Road_Segment::normal (double along, double from_center) const
{
  assert (mp_elevation_curve != 0);

  Two_Vector norm = mp_elevation_curve->normal (along + m_start_distance);
  double bank = deg_to_rad (mp_bank_curve->interpolate (along));
  double diff = from_center - left_road_width (along);
  if (mp_left_kerb)
    bank -= mp_left_kerb->angle (along, diff);
  diff = -from_center - right_road_width (along);
  if (mp_right_kerb)
    bank += mp_right_kerb->angle (along, diff);

  double x = norm.x;
  double y = sin (bank);
  double z = cos (bank) * norm.y;
  double this_angle = m_start_angle + arc () * along / m_length;
  return Three_Vector (cos (this_angle) * x - sin (this_angle) * y,
					   sin (this_angle) * x + cos (this_angle) * y,
					   z);
}

const Material& 
Road_Segment::material_at (double along, double from_center)
{
  //!!TODO handle skew
  double left_kerb_width = 0.0;
  double right_kerb_width = 0.0;

  if ((mp_left_kerb != 0) && mp_left_kerb->on_kerb (along))
	{
	  left_kerb_width = mp_left_kerb->width ();
	}

  if ((mp_right_kerb != 0) && mp_right_kerb->on_kerb (along))
	{
	  right_kerb_width = mp_right_kerb->width ();
	}

  if (from_center > left_road_width (along) + left_kerb_width)
	{
	  return m_materials [1];
	}
  else if (from_center > left_road_width (along))
	{
	  return m_materials [2];
	}
  else if (from_center > -right_road_width (along))
	{
	  return m_materials [3];
	}
  else if (from_center > -right_road_width (along) - right_kerb_width)
	{
	  return m_materials [4];
	}
  else
	{
	  return m_materials [5];
	}
}


void 
Road_Segment::set_start (const Three_Vector& start_coords, 
                         double start_distance,
                         double start_angle,
                         double start_bank,
                         const std::vector <double>& texture_offsets)
{
  m_start_distance = start_distance;

  m_center.x = start_coords.x - m_radius * sin (start_angle);
  m_center.y = start_coords.y + m_radius * cos (start_angle);

  m_start_coords = start_coords;
  if (m_radius == 0.0)
	{
	  m_end_coords = 
		Three_Vector (start_coords.x + m_length * cos (start_angle),
					  start_coords.y + m_length * sin (start_angle),
					  0.0);
	}
  else
	{
      m_end_coords = 
		Three_Vector (m_center.x + m_radius * sin (start_angle + arc ()),
					  m_center.y - m_radius * cos (start_angle + arc ()),
					  0.0);
	}

  m_start_angle = start_angle;
  if (m_start_angle > 2.0 * pi)
	m_start_angle -= 2.0 * pi;
  else if (m_start_angle < 0.0)
	m_start_angle += 2.0 * pi;

  // Initialize bank stuff.
  m_start_bank = start_bank;

  m_bank_points.clear ();
  m_bank_points.insert (m_bank_points.begin (), 
						Two_Vector (0.0, m_start_bank));
  m_bank_points.
	push_back (Two_Vector (m_length, m_end_bank));

  delete mp_bank_curve;
  mp_bank_curve = new Spline (m_bank_points, 0.0, 0.0);
}

double 
Road_Segment::left_width (double dist) const
{
  return mp_left_width->interpolate (dist);
}

double 
Road_Segment::right_width (double dist) const
{
  return mp_right_width->interpolate (dist);
}

double 
Road_Segment::left_road_width (double dist) const
{
  return mp_left_road_width->interpolate (dist);
}

double 
Road_Segment::right_road_width (double dist) const
{
  return mp_right_road_width->interpolate (dist);
}

void 
Road_Segment::set_length (double length)
{
  scale_widths (length / m_length);
  m_length = length;
}

void 
Road_Segment::scale (double factor)
{
  assert (factor != 0);
  scale_widths (factor);
  m_length *= factor;
  m_radius *= factor;
}

void
Road_Segment::scale_widths (double factor)
{
  mp_left_width->scale (factor);
  mp_right_width->scale (factor);
  mp_left_road_width->scale (factor);
  mp_right_road_width->scale (factor);
}

double 
Road_Segment::skew_slope () const
{
  // theta = +/- 90 degrees, and skew == 0 must be handled specially.
  const double sign = Vamos_Geometry::sign (m_radius);
  double theta = sign * arc () / 2.0;
  assert (cos (theta) != 0.0);
  assert (m_start_skew != 0.0);

  double phi = theta - sign * atan (m_start_skew);
  return (cos (phi) != 0.0)
    ? 1.0 / (sin (theta) / tan (phi) - cos (theta))
    : -1.0 / cos (theta);
}

double solve_quadratic (double a, double b, double c, double solution)
{
  return (-b + Vamos_Geometry::sign (solution) * sqrt (b*b - 4*a*c)) / (2*a);
}

double 
Road_Segment::skew_offset (double x, double y) const
{
  if (m_start_skew == 0.0)
    return 0.0;
  if (cos (arc () / 2.0) == 0.0)
    return x - sqrt (m_radius * m_radius - y * y);

  double slope = skew_slope ();

  return solve_quadratic (slope * slope - 1.0,
                          -2.0 * (slope * std::abs (m_radius) + x),
                          m_radius * m_radius - x * x - y * y,
                          -m_start_skew * m_radius);
}

double 
Road_Segment::skewed_radius (double x, double y, double offset) const
{
  x -= offset;
  return sqrt (x*x + y*y);
}

double
Road_Segment::coordinates (const Three_Vector& world_pos,
						   Three_Vector& track_pos)
{
  if (m_radius != 0.0)
    {
      const double sign = Vamos_Geometry::sign (m_radius);
      const double r0 = sign * m_radius;
      const double half_angle = sign * arc () / 2.0;

      track_pos = (world_pos - m_center).
        rotate (0.0, 0.0, (sign * pi - arc ()) / 2.0 - m_start_angle);

      const double offset = -skew_offset (track_pos.x, track_pos.y);
      const double r1 = skewed_radius (track_pos.x, track_pos.y, offset);
      const double theta = 
        sign * (atan2 (track_pos.y, track_pos.x - offset)) + half_angle;

      track_pos.x = r0 * theta;
      const double delta_x = offset + r1 * cos (half_angle)
        - r0 * cos (half_angle);
      const double delta_y = (r1 - r0) * sin (half_angle);
      track_pos.y = sign * Three_Vector (delta_x, delta_y, 0.0).magnitude () 
        * cos (atan (m_start_skew));
      if (delta_x > 0.0)
        track_pos.y *= -1.0;
      track_pos.z = 0.0;
    }
  else
    {
      track_pos = (world_pos - m_center).rotate (0.0, 0.0, -m_start_angle);
      double extra = track_pos.y * (m_start_skew - m_end_skew);
      double slope = m_length / (m_length + extra);
      double offset = slope * m_start_skew * track_pos.y;
      track_pos.x = slope * track_pos.x + offset;
    }
  track_pos.z = elevation (track_pos.x, track_pos.y);
  if (track_pos.x < 0.0)
    return track_pos.x;
  if (track_pos.x > m_length)
    return track_pos.x - m_length;
  return 0.0;
} 

Three_Vector
Road_Segment::position (double along, double from_center) const
{
  double x = m_center.x;
  double y = m_center.y;
  if (m_radius == 0.0)
    {
      x += along * cos (m_start_angle) - from_center * sin (m_start_angle);
      y += along * sin (m_start_angle) + from_center * cos (m_start_angle);
    }
  else
    {
      double angle = m_start_angle + along / m_radius;
      x += (m_radius - from_center) * sin (angle);
      y -= (m_radius - from_center) * cos (angle);
    }

  return Three_Vector (x, y, elevation (along, from_center));
}
